(*
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

(* $Id: fociModelChecker_concrete.ml,v 1.6 2007/08/29 01:41:53 beyer Exp $
 *
 * This file is part of the BLAST Project.
 *)


(**
 * This module implements the foci model checking algorithm.
 *)

open Ast
open BlastArch

module Stats = Bstats
module C_System_Descr = BlastCSystemDescr.C_System_Descr
module C_Command = BlastCSystemDescr.C_Command
    
(* screw the functor stuff -- because its just not working!! *)
 

module Make_Foci_Model_Checker =
  functor(Abstraction : ABSTRACTION) ->
struct

  module Region = Abstraction.Region

  module Operation = Abstraction.Operation

  module Tree = BiDirectionalLabeledTree

  type error_path = (Region.t * (Operation.t list) * Region.t)


type marking_data = {
    mutable time_stamp : int ;  (* RJ: Ok, I think the time stamp is useless! *)
    mutable region     : Region.t ;
    (* mutable ?? *) lvalue_map : Abstraction.lvmap_t ;     
}
      
let print_marking_data fmt md =
  Format.fprintf fmt "@[time_stamp=%d;@ region=" md.time_stamp ;
  Region.print fmt md.region ;
  Format.fprintf fmt "@]"
    
	  
let print_marking fmt = function
  md ->
      Format.fprintf fmt "@[(" ;
        print_marking_data fmt md ;
      Format.fprintf fmt ")@]" 

type node_data = {
    id           : int ;
    mutable mark : marking_data ; 
  }
		   

let print_node_data fmt nd =
  Format.fprintf fmt "@[Data(@[id=%d;@ mark=" nd.id ;
  print_marking fmt nd.mark ;
  Format.fprintf fmt "@])@]"


type tree_node = (node_data, (Operation.t * Ast.Predicate.predicate)) Tree.node 
(* Different from other trees *)
    
    
  (* helper functions to access a tree node's data *)
let get_node_data n = Tree.get_node_label n
    
let get_op n =
  match (Tree.get_parent_edge n) with
    Some(e) ->
      (Tree.get_edge_label e)
  | None ->
          invalid_arg "get_op: input node has no parent edge"
	
let get_parent n =
  match (Tree.get_parent_edge n) with
    Some(e) ->
      (Tree.get_source e)
  | None ->
      invalid_arg "get_parent: input node has no parent edge"
	
let get_children n = Tree.get_children n
    
let get_id n = (get_node_data n).id
    
let get_marking n = (get_node_data n).mark
    
let get_marking_data n = get_marking n
	  
let get_time_stamp n =
  try (get_marking_data n).time_stamp
  with Invalid_argument "get_marking: input node is unprocessed" ->
    invalid_arg "get_time_stamp: input node is unprocessed"
      
let print_tree_node fmt n =
  Format.fprintf fmt "@[TreeNode(@[label=" ;
  print_node_data fmt (get_node_data n) ;
  begin
    match (Tree.get_parent_edge n) with
      None ->
        Format.fprintf fmt ";@ no parent"
    | Some(e) ->
            Format.fprintf fmt ";@ incoming_op=" ;
        Operation.print fmt (fst (Tree.get_edge_label e)) ;
        Format.fprintf fmt ";@ parent_id=%d"
          (get_id (Tree.get_source e))
  end ;
  let children_ids = List.map get_id (Tree.get_children n)
  in
  Format.fprintf fmt ";@ children_ids=" ;
      Misc.list_printer_from_printer Format.pp_print_int fmt children_ids ;
  Format.fprintf fmt "@])@]"
    
    
let get_region n =
  try (get_marking_data n).region
  with Invalid_argument "get_marking: input node is unprocessed" ->
    invalid_arg "get_region: input node is unprocessed"
      

let get_lvalue_map n = 
  try (get_marking_data n).lvalue_map
  with Invalid_argument "get_marking: input node is unprocessed" ->
    invalid_arg "get_lvalue_map : input node is unprocessed"
    

  (* helper class to create nodes with appropriate ids *)
class tree_node_creator =
  object
  val mutable next_id = 0
			  
  val tree_node_table = Hashtbl.create 1009
			  (* hash int -> tree node for quick access to a tree node *)
			  (* q: why not an array instead ? -- as nodes get deleted ... This can be changed later *)
			  
  method create_root m =
    let this_id = next_id in
    next_id <- next_id + 1 ;
    let rv = ((Tree.create_root { id  = this_id ;
				  mark = m ;
				}) : tree_node) 
    in
    Hashtbl.add tree_node_table this_id (rv);
    rv
      (* FOCI *)
  method create_child m op par =
    let this_id = next_id in
    next_id <- next_id + 1 ;
    let rv = 
      ((Tree.create_child { id  = this_id ;
			    mark = m ;
			  } op par) : tree_node)
    in
      Hashtbl.add tree_node_table this_id (rv); 
       (* does this store only a pointer ... ? I think so ... but should check *)
      rv
	
  (* Jhala: It would be nice to have a destroy as well -- 
     or are we relying on some sort of Garbage collection to handle that ? *)
  method delete_children par = 
    let children = Tree.get_children (par : tree_node) in
    let child_ids = List.map get_id children in
    let _ = List.iter (Hashtbl.remove tree_node_table) child_ids in
      Tree.delete_children (par : tree_node)

  method node_from_id id = 
    try
      (Hashtbl.find tree_node_table id)
    with Not_found -> failwith "get_node_from_id: unknown id!"
      
end

exception RestartException
    
  (* statistic variables *)
let stats_nb_iterations = ref 0
let stats_nb_iterations_of_outer_loop = ref 0 (* added for races *)
let stats_nb_created_nodes = ref 0
let stats_nb_refinment_processes = ref 0
let stats_nb_refined_nodes = ref 0
let stats_nb_proof_tree_nodes = ref 0
let stats_nb_proof_tree_covered_nodes = ref 0
let stats_nb_deleted_nodes = ref 0
let local_restart_counter = ref 1

let reset_stats =
  stats_nb_iterations := 0 ;
  stats_nb_refinment_processes := 0 ;
  stats_nb_created_nodes := 0 ;
  stats_nb_refined_nodes := 0 ;
  stats_nb_proof_tree_nodes := 0;
  stats_nb_proof_tree_covered_nodes := 0
       
let print_stats fmt () =
  Format.fprintf fmt "@[<v>@[Nb iterations of outer while loop:@ %d@]" !stats_nb_iterations_of_outer_loop ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[<v>@[Nb iterations of reachability:@ %d@]" !stats_nb_iterations ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb created nodes:@ %d@]" !stats_nb_created_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refinment processes:@ %d@]" !stats_nb_refinment_processes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb refined nodes:@ %d@]" !stats_nb_refined_nodes ;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree nodes:@ %d@]" !stats_nb_proof_tree_nodes;
  Format.fprintf fmt "@ " ;
  Format.fprintf fmt "@[Nb proof tree covered nodes:@ %d@]@]" !stats_nb_proof_tree_covered_nodes;
  Format.fprintf fmt "@[Nb deleted nodes:@ %d@]@]" !stats_nb_deleted_nodes;
  ()

  (************************************************************************************************)
  (** Model checker data types **)
type model_check_outcome =
  | No_error of tree_node
  | Error_reached of (error_path * tree_node)

(* let global_variable_table = Hashtbl.create 2 *) (* To keep the abstraction interface happy.
						The global variable table was used by
						the TAR algorithm.
						*)

    (* the set of pending nodes to be processed *)
let unprocessed_nodes =  
  if (Options.getValueOfBool "bfs") then new Search.bfsiterator
  else new Search.dfsiterator
  
let cutoff_nodes = new Search.dfsiterator (* aka a stack *)


    (* helper object to create nodes *)
let tnc = new tree_node_creator

    (* the global time used for time stamps *)
let time = ref 0 

    (* the reached region is initially empty *)
let reached_region = ref Region.bot

let sig_caught_bit = ref false     

  (**************************************************************************************************)

type error_check =
    EmptyErrorAtNode of Region.t * tree_node * (tree_node list)
  | NonEmptyErrorAtRoot of Region.t * tree_node * (tree_node list)

 (*****************************************************************************************************************)
  (************************************  FOCI MODEL CHECKER loop ***************************************************) 
  (*****************************************************************************************************************)

 

let rec reclaim n = 
  if (Tree.has_child n || (not (Tree.has_parent n))) then ()
  else 
    begin
      let par = get_parent n in
	Tree.delete_child n; 
	reclaim par
    end
    

(* covering data structures *)

(* the "reach set" *)
let cov = Region.emptyCoverers ()

(*  define maps to track who is covered by whom -- 
    am postponing actually using these data structures *)

(* int -> int list map -- list of nodes that i covers *)
let covers_map = Hashtbl.create 1009

(* int -> int map -- the node that covers i *)
let coveredby_map = Hashtbl.create 1009

let cover_maps_add coverer coveree = 
  let [coverer_id;coveree_id] = List.map get_id [coverer;coveree] in
    Hashtbl.replace coveredby_map coveree_id coverer_id;
    Misc.hashtbl_check_update covers_map coverer_id coveree_id
         
let cover_maps_delete coverer coveree = 
  let [coverer_id;coveree_id] = List.map get_id [coverer;coveree] in
    Hashtbl.remove coveredby_map coveree_id;
    Misc.hashtbl_delete covers_map coverer_id coveree_id
    

(* This adds the successors of a node to the tree *)

let add_children n =
  let n_region = get_region n in
  let n_lvalue_map = get_lvalue_map n in
  let add_child op = 
    let 
      (new_region,new_lvalue_map,new_edge_constraint) = 
      Abstraction.foci_post (n_region,n_lvalue_map) op 
    in
    let n_op_md = 
      { time_stamp = (time := !time + 1 ; !time) ;
        region = new_region; 
	lvalue_map = new_lvalue_map;
      }
    in
      (tnc#create_child (n_op_md) (op,new_edge_constraint) n)
  in
    List.map add_child (Abstraction.enabled_ops n_region)


(* this is the core refine function -- used both when:
   1. A counterexample is found,
   2. For covered checking.
*)

exception FeasiblePath
(* this raises the exception FeasiblePath *)
let update_path_nodes n final_pred =
  let _ = Message.msg_string Message.Debug "In update_path_nodes" in
  let path = Tree.path_to_root n in
  let _reg_l = List.map get_region path in
    (* inject the condition we want into the last region *)
  let last_reg::prev = List.rev _reg_l in
  let new_last = Abstraction.insert_data_reg last_reg 
		   (Abstraction.pred_to_data_region final_pred)
  in
  if not (final_pred = Ast.Predicate.False) then (* no update reqd. if last reg is empty *)
  begin
  let reg_l = List.rev (new_last::prev) in
    (* now we have the list of regions *) 
  let (op_l , block_l) = Misc.unzip (List.map (get_op) ( List.tl path)) in
  let (ref_start_pos, new_reg_l,_) = 
    let reg_array = Array.of_list reg_l in
    let op_array = Array.of_list op_l in
    let block_array = Array.of_list block_l in
    let lvmap_array = Array.of_list (List.map get_lvalue_map path) in
      Abstraction.fociModelChecker_refine_trace reg_array op_array block_array (Some lvmap_array) 
  in
  let _ = Message.msg_string Message.Debug "Back from fmc_refine_trace" in
    (* this new_reg_l has length one less than path as the thing for the root node is just true *)
  let update_node node reg = (* we will do the "intersection" here *)
    let _ = Message.msg_string Message.Debug "Updating node:" in
    let old_reg = get_region node in
    let new_reg = Region.cap old_reg reg in
    Message.msg_string Message.Debug 
	      (Printf.sprintf "Old : %s" (Region.to_string old_reg));
    Message.msg_string Message.Debug
	      (Printf.sprintf "New : %s" (Region.to_string new_reg));
      (get_node_data node).mark <- {time_stamp = 0; 
				    region = new_reg; 
				    lvalue_map = (get_lvalue_map node)} 
  in
    if (ref_start_pos = -1) then raise FeasiblePath
    else
    (try
       List.iter2 update_node (List.tl path) (List.tl new_reg_l)
     with e -> 
       Message.msg_string Message.Error "focimodelchecker_concrete: Failure i update_node !"; 
       raise e)
     end        
	


let ac_is_covered cover_check_fun n = 
  let _ = Message.msg_string Message.Debug "In ac_is_covered" in
  let n_reg = get_region n in
    if (Region.is_empty n_reg) then true
    else
      begin
	let candidates_indices = Region.findCoverCandidates cov (get_region n) in
	let candidates = List.map (tnc#node_from_id) candidates_indices in
	let rec cover_process l = 
	  match l with
	      [] -> false
	    | h::t -> if cover_check_fun n h then true else cover_process t
	in
	  cover_process candidates
      end

(* check whether the node n is covered -- finds the possible candidates, using cov, and then 
   checks for each candidate using the cover_check_and_update *)
    
let is_covered n =        
  let _ = Message.msg_string Message.Debug "In is_covered" in
  let cover_check_and_update node cand_node = 
    (* return true if cover success ... in which case all updating is done, false o.w. *)
    let _ = Message.msg_string Message.Debug "In cover_check_and_update" in
    try 
      let cand_reg = get_region cand_node in
      let target_pred = Ast.Predicate.negate (Abstraction.extract_pred cand_reg) in
      let _ = Message.msg_string Message.Debug ("Candidate region: "^(Ast.Predicate.toString target_pred)) 
      in
	update_path_nodes node target_pred;
        let n_data_mark = (get_node_data n).mark in 
        (get_node_data n).mark  <- {n_data_mark with region = cand_reg};
	true 
        (* the path is not feasible, 
           the intermediate nodes get updated,
           but leave the last node region = cover candidate,
           node is covered ... return true ! *)
    with FeasiblePath -> false
  in
    ac_is_covered cover_check_and_update n

let is_covered_noupdate n = 
  let _ = Message.msg_string Message.Debug "In is_covered_noupdate" in
  let cover_check node cand_node = 
    (* returns true if reg.node \leq reg.cand_node -- doesn't change the tree at all *)
    let _ = Message.msg_string Message.Debug "In cover_check" in
      Region.leq (get_region node) (get_region cand_node) 
	(* It may be better to union all the candidates and check if this node is contained
	   in the union -- but if the theory is "convex" maybe this doesnt make a difference ?*)
  in
    ac_is_covered cover_check n

let output_path n = 
  let path = Tree.path_to_root n in
  let (op_l , block_l) = Misc.unzip (List.map (get_op) ( List.tl path)) in
  let print_op x = 
    let (l1,l2,rv) = Abstraction.op_to_string x in
    Message.msg_string Message.Normal (Printf.sprintf "%d :: %s :: %d \n\n" l1 rv l2) 
 in
  List.iter print_op op_l

let print_tree_from_root root =
  let nl_to_string nl = 
    let (regs_l,loc_s) = Abstraction.reg_to_strings (nl.mark.region) in
    let id = nl.id in
    Printf.sprintf "%d $$ (%s,%s)" id (loc_s) (Misc.strList regs_l)
  in
  let el_to_string (el,_) =
    let (_,_,s) = Abstraction.op_to_string el in
        s
  in
  Message.msg_string Message.Normal "**************************************************";
  Message.msg_string Message.Normal "**** Printing the Reachability Tree **************";
  Message.msg_string Message.Normal "**************************************************";
  let node_iter n = 
    Message.msg_printer Message.Normal print_tree_node n
  in
    Tree.iter_descendants_nodes node_iter root;
    let name = "reachtree" in
    Tree.output_tree_dot nl_to_string el_to_string (name^".dot") [root];
    Sys.command (Printf.sprintf "dot -Tps -o %s.ps %s.dot" name name);

exception PostCheckException of (tree_node * tree_node)

exception CoveredCheckException of tree_node

let check_tree root = 
  (* first, just check that each edge, is good -- then we have to check the leaves *)
  let phi_table = Hashtbl.create 1 in
  let check_post (n1, (op,_) , n2) = 
    let _ = Message.msg_string Message.Debug ("Checking Node: "^(string_of_int (get_id n2))) in
    let ac_post_reg = Abstraction.post (get_region n1) op phi_table in
    let check1 =       
      if Region.leq ac_post_reg (get_region n2) then ()
      else raise (PostCheckException (n1,n2))
    in
    let check2 = 
      if BiDirectionalLabeledTree.has_child n2 then () 
      else 
	if (is_covered_noupdate n2) then () 
	else raise (CoveredCheckException n2)
    in  ()
  in
    try
      BiDirectionalLabeledTree.iter_descendants_edges check_post root
    with
	CoveredCheckException n -> 
	  begin
	    Message.msg_string Message.Error "Covered Check Fails!!";
	    Message.msg_printer Message.Error print_tree_node n
	  end
      | PostCheckException (n1,n2) ->
	  begin
	    Message.msg_string Message.Error "Post Check Fails!!";
	    Message.msg_string Message.Error "Source Node:";
	    Message.msg_printer Message.Error print_tree_node n1;
	    Message.msg_string Message.Error "Target Node:";
	    Message.msg_printer Message.Error print_tree_node n2
	  end

let model_check init err = 
  (* random helper functions go here *)
  let tree_root =
    tnc#create_root { time_stamp = !time; region = init; lvalue_map = Abstraction.new_lvmap_t 0}
  in
    unprocessed_nodes#add_element (tree_root) ;
    let rec mc_iter () = 
      while (unprocessed_nodes#has_next ()) do
      (* begin debug info *)
      stats_nb_iterations := !stats_nb_iterations + 1;
      if (!stats_nb_iterations mod 100 == 0) then
	begin
          Message.msg_string Message.Normal ("While loop tick:"^(string_of_int !stats_nb_iterations));
	  flush stdout; flush stderr;
	end ;
      Message.msg_string Message.Debug 
	"\n\n****************************************************************************\n" ;
      Message.msg_string Message.Debug "Next iteration of foci-model-check's big while-loop" ;
      Message.msg_printer Message.Debug Format.pp_print_int (!stats_nb_iterations);
      (* end debug info *)
      (* let's process the next pending node *)
      let n = unprocessed_nodes#next () in
      let _ =
	Message.msg_string Message.Debug (*Debug Normal*) "Now processing tree node:" ;
	Message.msg_printer Message.Debug (*Debug Normal*) print_tree_node n 
      in
      let n_region = get_region n in
      let err_n = Stats.time "cap" (Region.cap n_region) err in
	if (Region.is_empty err_n)
	then
          (* no error found at this node *)
	  if is_covered n then 
	    begin
	      Message.msg_string Message.Debug ("Covered node: "^(string_of_int (get_id n)));
	      cutoff_nodes#add_element n
	    end
	  else 
	    begin
	      Message.msg_string Message.Debug "Not covered -- add children etc.";
	      Region.addCoverer cov (get_region n) (get_id n);
	      unprocessed_nodes#add_list (add_children n) 
	    end
	else
	  try
	    Message.msg_string Message.Debug "Hit error: update";
            update_path_nodes n Ast.Predicate.True
	  with FeasiblePath -> 
                begin
            Message.msg_string Message.Error "Real Error Found! :-(";
             output_path n;
            raise FeasiblePath
          end
    done;
    (* Now check if the cutoff nodes are really cut off -- if not, add them to the set of unprocessed nodes *)
    (* This check should be the side-effect free -- i.e. is there someone in the tree that covers them -- without
       inducing further refines along the way *)

    let filter_fun n = 
      let rv = 
	if is_covered_noupdate n then true else 
	  (* if the node isn't "directly" covered, then we put it in the 
	     worklist, and when we process it subsequently, we will check if 
	     it is "indirectly" covered *)
	  begin
	    unprocessed_nodes#add_element n;
	    false
	  end
      in
      let _ = Message.msg_string Message.Debug (Printf.sprintf "Really covered check: %d %b " (get_id n) rv) in
	rv
    in
      cutoff_nodes#filter filter_fun;
      if unprocessed_nodes#has_next () then mc_iter ()
      else ()
  in
    try
      mc_iter ();
      if (Options.getValueOfBool "showtree") then 
	begin
	  print_tree_from_root tree_root;
	  check_tree tree_root
	end;
	Message.msg_string Message.Normal "Program safe. Error not reachable! :-)";
	Message.msg_string Message.Error "Program safe. Error not reachable! :-)";
    with FeasiblePath -> 
              raise FeasiblePath
	(* RJ: Should return "Error_Reached etc. etc." *)
end
