(*  
BLAST is a tool for software model checking.
This file is part of BLAST.

Copyright (c) 2002-2007, The BLAST Team.
All rights reserved. 

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. Neither the name of the authors nor their organizations 
   may be used to endorse or promote products
   derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

(This is the Modified BSD License, see also
 http://www.opensource.org/licenses/bsd-license.php)

The BLAST Team consists of
        Dirk Beyer (SFU), Thomas A. Henzinger (EPFL),
        Ranjit Jhala (UCSD), and Rupak Majumdar (UCLA).

BLAST web page:
        http://mtc.epfl.ch/blast/

Bug reports:
        Dirk Beyer:      firstname.lastname@sfu.ca or
        Rupak Majumdar:  firstname@cs.ucla.edu or
        Ranjit Jhala:    lastname@cs.ucla.edu
 *)

module O = Options
module M = Message
module E = Ast.Expression
module P = Ast.Predicate
module AA = AliasAnalyzer
module Op = Abs_op.Operation
module U = Abs_util
module C_SD = BlastCSystemDescr.C_System_Descr

module LvalMap = Map.Make(struct
  type t = E.lval
  let compare lv1 lv2 = compare (E.lvalToString lv1) (E.lvalToString lv2)
end)

let lvm_cf_flag () = 
  O.getValueOfBool "cf" || O.getValueOfBool "enablerecursion" 

let get_new_frame =
  let frame = ref 0 in
  (fun () -> (Misc.inc frame; !frame))

let filter_copy f src dst =
  LvalMap.fold (fun lv d m -> if f lv then LvalMap.add lv d m else m) src dst

class lvmap = 
  object (self)
    val mutable cur = LvalMap.empty
    val mutable hist = []
    val mutable frames = [0]
 
    method print_frames () =
      M.msg_string M.Debug ("Frames: "^(Misc.strList (List.map (string_of_int) frames)))

    method print () =
      LvalMap.fold 
      (fun lv e () ->
        M.msg_string M.Debug 
        (Printf.sprintf "%s ---> %s" (E.lvalToString lv) (E.toString e))) 
      cur ()
   
    method clear () = 
      cur <- LvalMap.empty;
      hist <- [];
      frames <- [0]

    method update lv = 
      let lv' = 
        if O.getValueOfBool "clock" then U.getNewTmpVar_clock lv 
        else U.getNewTmpVar lv in
      cur <- LvalMap.add lv lv' cur

    method mem lv = LvalMap.mem lv cur

    (* TBD:WHY? we need this to be deterministic ? is it to do with "clock" ? *)
    method private map_lv lv =
      if E.is_interpreted_lval lv then E.Lval lv else 
        try LvalMap.find lv cur with Not_found -> 
          let id =
            try (string_of_int (List.hd frames)) with _ -> failwith "BAD FRAMES!" in
          let (id,cand) = 
            if (O.getValueOfBool "timezero") then (id,[lv])
            else ((E.lvalToString lv)^id,lv::(AA.ac_get_lval_aliases true lv)) in
          let lvs = List.filter (fun lv' -> not (self#mem lv')) cand in
          let vals = List.map (fun lv' -> E.Chlval (lv',id)) lvs in
          let _ = cur <- List.fold_right2 LvalMap.add lvs vals cur in
          E.Chlval (lv,id) 

    method map_expr e = 
      E.address_normalize (E.deep_alpha_convert (self#map_lv) e)

    method map_pred p = 
      P.address_normalize (P.deep_alpha_convert (self#map_lv) p) 

    method push () =
      if lvm_cf_flag () then
        (hist <- cur::hist;
         frames <- (get_new_frame ())::frames;
         cur <- filter_copy C_SD.is_global cur LvalMap.empty)
   
    method pop () = 
      if lvm_cf_flag () then
        match (hist,frames) with (h::hist', _::frames') -> 
          (cur <- filter_copy C_SD.is_global cur h;
           hist <- hist';
           frames <- frames')
        | _ -> failwith "error: lvmap.pop"

    method init start ops =
      self#clear ();
      if (lvm_cf_flag ()) then
        let stack = 
          List.fold_left 
          (fun stk op -> 
            match Op.get_info op with
              Op.Ret _ -> (try List.tl stk with _ -> failwith "lvmap.init: mismatched call/ret")
            | Op.Normal -> stk 
            | Op.Call (_, _, f,_) when (C_SD.isAsyncCall op) -> stk
            | Op.Call (_, _, f,_) -> f :: stk) [start] ops in
        List.iter (fun _ -> self#push ()) stack 
end
