(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
open VampyrePretty
module E = VampyreErrormsg


(* This defines data types for the output of the parser. *)
type var   = string
let  count = ref 0 
let  newVar name = 
  incr count;
  name ^ "_" ^ (string_of_int !count)
let  resetVar = count := 0
let  varName v = 
  try
    String.sub v 0 (1 + String.rindex v '_')
  with Not_found -> E.s (E.bug "%s is not a variable name" v)
  

(* predicates *)
type pred =
    True			(* true       *)
  | False			(* false      *)
      
  | All   of var * pred	        (* forall v.P *)
  | Exist of var * pred	        (* exist v.P  *)
  | And  of pred * pred	        (* P /\ Q     *)
  | Or   of pred * pred	        (* P \/ Q     *)
  | Not of pred			(* ~ P        *)
  | Imp  of pred * pred	        (* P => Q     *)

  | Comment of string * pred	(* attach a comment to any predicate *)
  | Atom of atom

and atom = 
  | Beq of atom * atom           (* [Jhala] *)  
  | Eq of exp * exp             (* atoms *)
  | Neq of exp * exp
  | Geq of exp * exp
  | Gt of exp * exp
  | Leq of exp * exp
  | Lt of exp * exp

  | BFun  of string * exp list		(*   all other boolean functions   *)
  | NotBFun of string * exp list        (* negated literals *)

and exp =                               (* Expressions *)
    Var of var                          (* Variables *)
  | Int of int                          (* Constants *)
  | Plus of exp * exp
  | Minus of exp * exp
  | Times of exp * exp

                                        (* Some expressions for modular 
                                         * arithmetic *)
  | Div   of int * exp                  (*   division by int  *)
  | Mod   of int * exp                  (*   modulo by int    *)
  | Floor of int * exp                  (*   floor to int     *)
  | Ceil  of int * exp                  (*   ceiling to int   *)
(*  | Eatom of atom *)                       (* [Jhala] *) 
  | Fun   of string * exp list		(*   all other non-bool functions  *)
    
type input =
    Goal   of string * pred
  | Axiom  of string * var list * pred list
  | Import of string
  | Decls  of (string * string list) list


(********* VampyrePretty printing ***********)

let getexpprec = function
    (Var _ | Int _ | Fun _ | Floor _ | Ceil _) -> 3
  | (Times _ | Mod _ | Div _) -> 2
  | (Plus _ | Minus _) -> 1

let rec getpredprec = function
    (True | False | Atom(BFun _) | Atom(NotBFun _)) -> 3
  | Atom _ -> 2
  | Not _ -> 2
  | And _ -> 1
  | (Or _ | Imp _) -> 0
  | (All _ | Exist _) -> -1
  | Comment(_, p) -> getpredprec p

let printLFSyntax = ref false

                                        (* pretty print a binary expression. 
                                         * Parenthesize the subexpressions if 
                                         * necessary *)
let binop name prectop precl dl precr dr = 
  if !printLFSyntax then
    dprintf "(%s @[%a@!%a@])" name insert dl insert dr
  else
  dprintf "@[%a %s@?%a@]"
    insert (if prectop > precl then dprintf "(%a)" insert dl else dl)
    name
    insert (if prectop > precr then dprintf "(%a)" insert dr else dr)
    
let dv () v = text v


let rec de () e = 
  match e with 
    Int i -> num i
  | Var v -> dv () v
  | Plus(e1,e2) -> expbinop "+" (getexpprec e) e1 e2
  | Minus(e1,e2) -> expbinop "-" (getexpprec e) e1 e2
  | Times(e1,e2) -> expbinop "*" (getexpprec e) e1 e2
	(* [weimer] I want certainty 
	| Plus(e1,e2) -> dprintf "(%a + %a)" de e1 de e2
	| Minus(e1,e2) -> dprintf "(%a - %a)" de e1 de e2
	| Times(e1,e2) -> dprintf "(%a * %a)" de e1 de e2
	*)
        
  | Fun(op, []) -> text op
  | Fun(op, args) -> 
      if !printLFSyntax then
        dprintf "(%s @[%a@])" op (docList break (de ())) args 
      else
      dprintf "%s(@[%a@])" op (docList (chr ',' ++ break) (de ())) args
	| Mod(i, e1) -> dprintf "((%a) mod %d)" de e1 i
	| Div(i, e1) -> dprintf "((%a) div %d)" de e1 i
	| Floor(i, e1) -> dprintf "((%a) floor %d)" de e1 i
	| Ceil(i, e1) -> dprintf "((%a) ceil %d)" de e1 i
  | _ -> E.s (E.unimp "docExp")
  
and expbinop name prectop e1 e2 = 
  binop name prectop (getexpprec e1) (de () e1)  (getexpprec e2) (de () e2) 


let rec dp () p = 
  match p with 
    True -> text "true"
  | False -> text "false"
  | And(p1,p2) -> predbinop "/\\" (getpredprec p) p1 p2
  | Or(p1,p2) -> predbinop "\\/" (getpredprec p) p1 p2
  | Not(p1) -> dprintf "~ (%a)" dp p1 
  | Imp(p1,p2) -> predbinop "=>" (getpredprec p) p1 p2

  | All(v,p) -> 
      if !printLFSyntax then
        dprintf "@[(all [%a : exp]@!@]  %a)" dv v dp p
      else
        dprintf "@[all %a.@!@]  %a" dv v dp p

  | Exist(v,p) -> 
      if !printLFSyntax then
        dprintf "@[(exists [%a : exp]@!@]  %a)" dv v dp p
      else
        dprintf "@[exists %a.@!@]  %a" dv v dp p

  | Atom a -> da () a
  | Comment(s,p) -> dprintf "%% %s@! %a" s dp p

and da () a = 
  match a with
  | Beq(a1,a2) -> predbinop "==" (getpredprec (Atom a)) (Atom a1) (Atom a2)
  | Eq(e1,e2) -> expbinop "=" (getpredprec (Atom a)) e1 e2
  | Neq(e1,e2) -> expbinop "<>" (getpredprec (Atom a)) e1 e2
  | Geq(e1,e2) -> expbinop ">=" (getpredprec (Atom a)) e1 e2
  | Gt(e1,e2) -> expbinop ">" (getpredprec (Atom a)) e1 e2
  | Leq(e1,e2) -> expbinop "<=" (getpredprec (Atom a)) e1 e2
  | Lt(e1,e2) -> expbinop "<" (getpredprec (Atom a)) e1 e2

  | BFun(op, []) -> text op
  | BFun(op, args) -> 
      if !printLFSyntax then
        dprintf "(%s @[%a@])" op (docList (chr ',' ++ break) (de ())) args
      else
      dprintf "%s(@[%a@])" op (docList (chr ',' ++ break) (de ())) args
  | NotBFun(op, args) -> 
      dprintf "(not %a)" da (BFun(op,args))


and predbinop name prectop p1 p2 = 
  binop name prectop (getpredprec p1) (dp () p1)  (getpredprec p2) (dp () p2) 

let di () = function
    Goal(name, p) -> dprintf "%s : %a" name dp p
  | _ -> E.s (E.unimp "docInput")


(* Negate an atom *)
let negate = function
    Eq(e1,e2) -> Neq(e1,e2)
  | Neq(e1,e2) -> Eq(e1,e2)
  | Leq(e1,e2) -> Gt(e1,e2)
  | Lt(e1,e2) -> Geq(e1,e2)
  | Geq(e1,e2) -> Lt(e1,e2)
  | Gt(e1,e2) -> Leq(e1,e2)
  | BFun(n,args) -> NotBFun(n,args)
  | NotBFun(n,args) -> BFun(n,args)

(* Equivalence checking *)
let rec samee e e' = 
  match e, e' with
    (Int i), (Int i') -> i = i'
  | (Var v), (Var v') -> v = v'
  | Plus(e1,e2), Plus(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Minus(e1,e2), Minus(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Times(e1,e2), Times(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Div(n,e), Div(n',e') -> n=n' && samee e e'
  | Mod(n,e), Mod(n',e') -> n=n' && samee e e'
  | Floor(n,e), Floor(n',e') -> n=n' && samee e e'
  | Ceil(n,e), Ceil(n',e') -> n=n' && samee e e'
  | Fun(op,args), Fun(op',args') -> 
      op = op' &&
      List.fold_left2 (fun acc e e' -> acc && samee e e') true args args'
  | _ -> false

let rec samea a a' = 
  match a, a' with
    Eq(e1,e2), Eq(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Neq(e1,e2), Neq(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Geq(e1,e2), Geq(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Gt(e1,e2), Gt(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Lt(e1,e2), Lt(e1',e2') -> samee e1 e1' && samee e2 e2'
  | Leq(e1,e2), Leq(e1',e2') -> samee e1 e1' && samee e2 e2'
  | BFun(op,args), BFun(op',args') -> 
      op = op' &&
      List.fold_left2 (fun acc e e' -> acc && samee e e') true args args'
  | NotBFun(op,args), NotBFun(op',args') -> 
      op = op' &&
      List.fold_left2 (fun acc e e' -> acc && samee e e') true args args'
  | _ -> false


  (* The samep function does NOT do alpha-equivalence, as it should. But we 
   * hope that since the parser creates brand-new variables names, this will 
   * not be a problem *)
let rec samep p p' = 
  match p, p' with
    True, True -> true
  | False, False -> true
  | And(p1,p2), And(p1',p2') -> samep p1 p1' && samep p2 p2'
  | Or(p1,p2), Or(p1',p2') -> samep p1 p1' && samep p2 p2'
  | Not(p1), Not(p1') -> samep p1 p1'
  | Imp(p1,p2), Imp(p1',p2') -> samep p1 p1' && samep p2 p2'
  | All(v,p), All(v',p') -> v = v' && samep p p'
  | Exist(v,p), Exist(v',p') -> v = v' && samep p p'
  | Atom a, Atom a' -> samea a a'
  | _ -> false


let rec pushNeg p =
  match p with
    Not (p') ->
      begin
        match p' with
          True -> False
        | False -> True
	| And(p1, p2) -> Or(pushNeg (Not p1), pushNeg (Not p2))
	| Or(p1, p2) -> And(pushNeg (Not p1), pushNeg (Not p2))
	| Imp(p1, p2) -> And(pushNeg p1, pushNeg (Not p2))
	| Not(p1) -> pushNeg p1
	| All _ -> failwith "Do not push Negation through All"
	| Exist _ -> failwith "Do not push Negation through Exist"
	| Atom a -> Atom (negate a)
      end
  | True -> True
  | False -> False
  | And(p1, p2) -> And(pushNeg p1, pushNeg p2)
  | Or(p1, p2) -> Or(pushNeg p1, pushNeg p2)
  | Imp(p1, p2) -> Imp(pushNeg p1, pushNeg p2)
  | All (v,p1) -> All (v, pushNeg p1) 
  | Exist (v,p1) -> Exist (v, pushNeg p1) 
  | Atom a -> Atom a
