%{
module I = Input
module E = VampyreErrormsg
open VampyrePretty

let  parse_error msg =
    E.s (E.error "Parse error at %s: %s" 
           (E.getLineCol !E.current (symbol_start ())) msg)

let bindings : (string * I.var) list ref = ref [] 
let pushBinding vname = 
  let v = I.newVar vname in
  bindings := (vname, v) :: !bindings;
  v

let rec popBindings n res = 
  if n = 0 then res else
  begin
    (match !bindings with
      [] -> E.s (E.bug "no binding to pop")
    | _ :: rest -> bindings := rest);
    popBindings (n - 1) res
  end

let lookupVar name = 
  try
    let (_, v) = List.find (fun (n,_) -> n = name) !bindings in
    v
  with Not_found -> 
    parse_error ("Variable " ^ name ^ " not bound")

%}


%token <string> Id Var Use Comment
%token <int> Num
%token EXP PRED
%token PF
%token LPAREN RPAREN LBRACKET RBRACKET LBRACE RBRACE
%token DOT COLON ARROW
%token PLUS MINUS TIMES
%token MOD DIV FLOOR CEIL
%token LT GT LEQ GEQ EQ NEQ
%token TRUE FALSE AND OR NOT IMPL ALL EXISTS
%token BEGINDECL ENDDECL
%token EOF

%start main

%type <Input.input list> main
%type <Input.input> input
%type <Input.var list> axvars
%type <Input.pred list> axiom
%type <Input.pred> pred
%type <Input.exp> exp
%type <Input.exp list> exps
%type <Input.var> bind
%type <(string * string list) list> decls
%type <string list> decltype

%%

main:
    EOF					{ [] }
  | input main				{ $1 :: $2 }
;
input:
    Var COLON PF pred DOT		{ Input.Goal   ($1, $4)     }
  | Id COLON axvars axiom DOT		{ popBindings (List.length $3)
                                                      (Input.Axiom ($1, $3, 
                                                                    $4)) }
  | Use					{ Input.Import $1           }
  | BEGINDECL decls ENDDECL		{ Input.Decls  $2           }
;
axvars:
    					{ [] }
  | LBRACE Var COLON EXP  RBRACE axvars	{ (pushBinding $2) :: $6 }
  | LBRACE Var COLON PRED RBRACE axvars	{ (pushBinding $2) :: $6 }
;
axiom:
    PF pred				{ [$2] }
  | PF pred ARROW axiom		        { $2 :: $4 }
;
pred:
    TRUE				{ I.True  }
  | FALSE				{ I.False }
  | ALL bind pred			{ popBindings 1 (I.All  ($2, $3)) }
  | ALL LPAREN bind pred RPAREN 	{ popBindings 1 (I.All  ($3, $4)) }
  | EXISTS bind pred			{ popBindings 1 (I.Exist ($2, $3)) }
  | EXISTS LPAREN bind pred RPAREN 	{ popBindings 1 (I.Exist ($3, $4)) }
  | AND pred pred			{ I.And  ($2, $3) }
  | OR pred pred			{ I.Or   ($2, $3) }
  | NOT pred				{ I.Not ($2) }
  | IMPL pred pred			{ I.Imp  ($2, $3) }
  | LPAREN pred RPAREN			{ $2 }
  | Comment pred			{ I.Comment ($1, $2) }

  | EQ  exp exp				{ I.Atom(I.Eq($2, $3)) }
  | NEQ exp exp				{ I.Atom(I.Neq($2, $3)) }
  | GEQ exp exp				{ I.Atom(I.Geq($2,$3)) }
  | GT  exp exp				{ I.Atom(I.Gt($2, $3)) }
  | LEQ exp exp				{ I.Atom(I.Leq($2, $3)) }
  | LT  exp exp				{ I.Atom(I.Lt($2, $3)) }
  | Id					{ I.Atom(I.BFun($1, [])) }
  | LPAREN Id exps RPAREN		{ I.Atom(I.BFun($2, $3)) }
;


exp:
    Var					{ I.Var (lookupVar $1) }
  | Num					{ I.Int $1 }
  | PLUS  exp exp			{ I.Plus($2, $3) }
  | MINUS exp exp			{ I.Minus($2, $3) }
  | TIMES exp exp			{ I.Times($2, $3) }
  | Id					{ I.Fun($1, []) }
  | LPAREN Id exps RPAREN		{ I.Fun($2, $3) }
  | DIV   exp Num			{ I.Div($3, $2) }
  | MOD   exp Num			{ I.Mod($3, $2) }
  | FLOOR exp Num			{ I.Floor($3, $2) }
  | CEIL  exp Num			{ I.Ceil($3, $2) }
  | LPAREN exp RPAREN			{ $2 }
;
exps:
    exp					{ [$1] }
  | exp exps				{ $1 :: $2 }
;
bind:
    LBRACKET Var COLON EXP RBRACKET	{ pushBinding $2 }
;

decls:
    					{ [] }
  | Id COLON decltype DOT decls		{ ($1, $3) :: $5 }
;
decltype:
    EXP					{ ["exp"] }
  | EXP ARROW decltype			{ "exp" :: $3 }
  | PRED				{ ["pred"] }
  | PRED ARROW decltype			{ "pred" :: $3 }
;
