(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* Wes Weimer
 * CS 294-4, Spring '00
 *
 * Subgoal Manager
 *
 * General Theory: we receive as input a bunch of (subgoal,priority,proof)
 * tuples.  Eventually, when nothings else works, the prover asks us for the
 * next subgoal. 
 *
 *)

(* See Classnotes for 3/23/00 on Subgoal manager information *)

module E = VampyreErrormsg
module I = Input
module P = Proof
module S = Satproc
open VampyrePretty
open Dag

type subgoal = {
    owner			: Satproc.ident;	(* who is responsible for this subgoal *)
    goal		  : Input.pred list;		(* the subgoals G to be proved *)
    (* see the end of this file for instructions on getting an Input.pred
     * out of a standard DAG node
     *)
    proof_gen : Proof.proof list -> Proof.proof ; (* Prf[G] -> Prf[False] *)
      (* the responsible satproc must be willing to provide a proof that
       * proving G is sufficient *)
      (* your promise to the subgoal manager: "If someone can prove G, I
       can use G [and my current hypotheses] to prove falsehood." 
       *)
       priority  : int		  (* should we try this early? low means try early *)
       (* 0 means average, >0 means try later *)
       } ;;

(* 
 * a stack based structure for snapshots and undos 
 *)
type subgoalElem = 
		Snapshot
	| Subgoals of (subgoal list * (* yet to try *)
							   subgoal list ) (* already tried since last Undo *)
;;

let subgoalStack : subgoalElem list ref = ref [Subgoals([],[])] ;;

(* 
 * The Snapshot Completeness Bug
 *
 * 			|
 * 			| subgoals = {S,T}
 * 		 ==>
 * 		 / \
 * 		/   \
 * 	 /A    \B
 *
 * Assume we use S and T at A. Then they must still be available at B.
 * If we have an "assumption" everytime we try to prove a leaf (literal)
 * this won't be a problem, because we'll be marking our only local copies
 * as "tried". If not, we'd have to add code at every ==> to make copies of
 * all of the subgoals.
 *)

(* 
 * Record a snapshot 
 * copy all of the current subgoals
 *)
let snapshot () =
	match !subgoalStack with
		Subgoals(l) :: _ -> 
		  (* copy over all of the untried subgoals *)
			subgoalStack := Subgoals(l) :: Snapshot :: !subgoalStack
	| _ -> E.s (E.bug "SG snapshot error")
;;

(* 
 * Discard all assertions since last snapshot 
 *)
let undo () = 
	let rec loop = function
			Snapshot :: rest -> subgoalStack := rest
		| _ :: rest -> loop rest
		| [] -> E.s (E.bug "SG tried to pop an empty stack (too many undos?)")
	in 
		loop !subgoalStack
;;

(*
 * subgoal comparison function
 *)
let sgcompare a b = (a.priority <= b.priority) 
;;
	
(*
 * Add a new subgoal.
 *)
let add sg =
	match !subgoalStack with
		Subgoals(yet,old) :: rest -> 
		  let rec already_tried sglist = (
				match sglist with
					a::rest -> (
						if (a.goal = sg.goal) then (
							(if (!E.debugFlag) then ignore
								(printf "SG: REJECTING duplicate @[%a@]\n" 
									 (docList break (I.dp ())) sg.goal)) ;
							yet
						) else (already_tried rest))
				| [] ->  (if (!E.debugFlag) then ignore
									(printf "SG: adding subgoal @[%a@]\n" 
									 (docList break (I.dp ())) sg.goal)) ;
								 sg::yet 
			) in
			subgoalStack := 
				Subgoals(Sort.list sgcompare (already_tried old),old) :: rest 
	| _ -> E.s (E.bug "SG add error")
;;

(*
 * returns the next subgoal, if one is available
 *)
let getNext () =
	match !subgoalStack with
		Subgoals(yet :: yetrest, old) :: rest ->
			subgoalStack := Subgoals(Sort.list sgcompare yetrest,yet::old) :: rest ;
			if (!E.debugFlag) then ignore
				(printf "SG: trying @[%a@]\n" (docList break (I.dp ())) yet.goal) ;
			Some(yet)
	| Subgoals([],[]) :: rest ->
		  None 
	| _ -> 
			E.s (E.bug "SG getNext error: no subgoal marker on stack")
;;

(*
 * are we out of subgoals?
 *)
let empty () =
	match !subgoalStack with
		Subgoals(sg :: sgrest,old) :: rest -> false
	| _ -> true
;;

(*
 * Turn a DAG node back into an input.atom/exp. This is useful if you need
 * an input.pred for use in a subgoal. For example, if you'd like to submit
 * A /\ B as a subgoal, where A and B are dagnodes, use this as your
 * input.pred:
 * 		(Input.And(Input.Atom(inputAtomOfDagNode A),
 * 							 Input.Atom(inputAtomOfDagNode B)))
 *)
let rec inputAtomOfDagNode dn = 
	match dn.op with
	  Eq -> (I.Eq(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| Neq -> (I.Neq(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| Geq -> (I.Geq(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| Gt -> (I.Gt(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| Leq -> (I.Leq(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| Lt -> (I.Lt(inputExpOfDagNode (List.hd dn.args), 
								inputExpOfDagNode (List.nth dn.args 1)))
	| BFun(s) -> (I.BFun(s, (List.map inputExpOfDagNode dn.args)))
	| _ -> E.s (E.bug "subgoal.inputAtomOfDagNode: what is %a" Dag.dn dn )
and inputExpOfDagNode dn =
	match dn.op with
	  Plus -> (I.Plus(inputExpOfDagNode (List.hd dn.args), 
								  inputExpOfDagNode (List.nth dn.args 1)))
	| Minus -> (I.Minus(inputExpOfDagNode (List.hd dn.args), 
								  inputExpOfDagNode (List.nth dn.args 1)))
	| Times -> (I.Times(inputExpOfDagNode (List.hd dn.args), 
								  inputExpOfDagNode (List.nth dn.args 1)))
	| Div(i) -> (I.Div(i, inputExpOfDagNode (List.hd dn.args)))
	| Mod(i) -> (I.Mod(i, inputExpOfDagNode (List.hd dn.args)))
	| Floor(i) -> (I.Floor(i, inputExpOfDagNode (List.hd dn.args)))
	| Ceil(i) -> (I.Ceil(i, inputExpOfDagNode (List.hd dn.args)))
	| Fun(s) -> (I.Fun(s, (List.map inputExpOfDagNode dn.args)))
	| Int(i) -> (I.Int(i))
	| Var(v) -> (I.Var(v))
	| _ -> E.s (E.bug "subgoal.inputExpOfDagNode: not an exp: %a" Dag.dn dn)
;;






