(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

(* Pretty printer *)

let debug = false
let fprintf x = Printf.fprintf stderr x

type doc = 
    Nil
  | Text     of string
  | Concat   of doc * doc
  | CText    of doc * string
  | Line     
  | Break
  | Align
  | Unalign


let nil  = Nil
let (++) d1 d2 = Concat (d1, d2)
let text s     = Text s
let num  i     = Text (string_of_int i)
let chr  c     = Text (String.make 1 c)
let align      = Align
let unalign    = Unalign
let line       = Line
let break      = Break


let nest n d   = text (String.make n ' ') ++ align ++ d ++ unalign

let  seq sep f dl = 
  let rec loop = function
      []     -> nil
    | h :: t -> sep ++ f h ++ loop t
  in
  (match dl with
    [] -> nil
  | h :: t -> f h ++ loop t)

let docArray f () a = 
  let len = Array.length a in
  if len = 0 then 
    nil
  else
    let rec loop i =
      if i >= len then nil else
      f i a.(i) ++ loop (i + 1)
    in
    f 0 a.(0) ++ loop 1
      
let docOpt delem () = function
    None -> text "None"
  | Some e -> text "Some(" ++ (delem () e) ++ chr ')'
  
let docList sep f () dl = seq sep f dl

let insert () d = d


(************* LAYOUT functions ********************)

type status = 
    Open                                (* We have not yet decided whether to 
                                         * take this break or not. Both the 
                                         * col field and   *)
  | Taken                               (* We are definitely taking this one *)
  | NotTaken                            (* We are definitely not talking this 
                                         * one *)
type brk = {         id     : int;
             mutable col    : int;
                   align    : int ref;
             mutable status : status; } 

type breakData = 
    Brk  of brk
  | Algn of int ref


let printData breaks = 
  let lastCol = ref 99999 in
  let printOne = function
      Brk b -> 
        if b.status = Open then begin
          fprintf  "B%d(%d,%d), " b.id b.col (! (b.align));
            lastCol := b.col
        end else
          ()
  | Algn ar -> 
        if !ar <= !lastCol then begin
          fprintf "A(%d), " (!ar);
          lastCol := !ar
        end else
          ()
  in
  List.iter printOne breaks



let fit start accString width = 
  let width = if width = 0 then 99999 else width in
                                        (* A flag to say that we shouldn't try 
                                         * to perform the expensive line 
                                         * breaking for a while *)
  let breakAllMode = ref false in
  let breakId = ref 0  in

                                        (* Next mark all the still Open 
                                         * breaks as not taken up to the 
                                         * align that we are taking 
                                         * (starting at the most recent 
                                         * break)  *)
  let rec markNotTaken align = function
      Brk ({status=Open;} as b) :: rest ->
        b.status <- NotTaken;
        markNotTaken align rest
    | Algn ar :: rest -> 
        if align == ar then () else markNotTaken align rest
    | _ :: rest -> markNotTaken align rest
    | [] -> ()  in

                                        (* Function to perform some breaks. 
                                         * Takes a stack of breaks in scope, 
                                         * the width and the current column. 
                                         * Marks some breaks and returns the 
                                         * new column *)
  let performBreaks breaks col =
    let oneBrk col = 
      let _ = 
        if debug then begin
          fprintf "Before oneBrk (col = %d): " col;
          printData breaks;
          fprintf "\n"
        end else 
          () in
      let maxDelta = ref 0 in
      let maxCol   = ref 0 in
      let maxBrk   : brk ref = ref {id=0;col=0;align=ref 0;status=Open;} in
      let rec findMaxDelta = function
          [] -> ()
        | Brk b :: rest when b.status = Open ->
            if !maxDelta < b.col - ! (b.align) then begin
              maxDelta := b.col - ! (b.align);
              maxCol   := b.col;      (* Create a private copy *)
              maxBrk   := b;
            end else
              ();
            findMaxDelta rest
        | _ :: rest -> findMaxDelta rest
      in
      let _ = findMaxDelta breaks in
      let _ = 
        if !maxDelta > 0 then 
                                        (* Adjust the positions of following 
                                         * breaks and aligns  *)
          let delta = !maxDelta in
          let align = !maxBrk.align in
                                        (* Now we can adjust all the column 
                                         * of all the Open breaks and of all 
                                         * the aligns until we hit the break 
                                         * we are taking. This is because the 
                                         * breaks are in reversed order *)
          let rec adjustColumns = function 
              Algn ar :: rest -> 
                ar := !ar - delta; 
                assert (!ar >= 0);
                adjustColumns rest
            | Brk ({status=Open;} as b) :: rest -> 
                if b == !maxBrk then begin
                  b.status <- Taken;
                  b.col <- b.col - delta;
                  markNotTaken align rest
                end else begin
                  b.col <- b.col - delta;  
                  adjustColumns rest
                end
            | _ :: rest -> adjustColumns rest
            | _ -> failwith "Pretty: cannot find the break"
          in
          if debug then fprintf "Adjusting for break at %d\n" (!maxCol) else();
          flush stderr;
          adjustColumns breaks
        else
          ()
      in
      begin 
        if debug then begin
          fprintf "  B=%d, D=%d: " (!maxCol) (!maxDelta);
          printData breaks;
          fprintf "\n"
        end else
          ();
        !maxDelta
      end
          (* oneBrk *)        
    in
                                        (* Iterate oneBrk while progress can 
                                         * be made, until necessary *)
    let rec loop col = 
      if col > width then
        let delta = oneBrk col in
        if delta > 0 then          (* If some progress was made *)
          begin
            if debug && col - delta > width then
              begin fprintf "  Retrying oneBrk\n"; flush stderr end
            else
              ();
            loop (col - delta)
          end
        else                            (* Cannot improve *)
          col
      else   
        col
    in
    loop col
    (* performBreaks *)
  in

                                        (* Collect here all the breaks in the 
                                         * order in which they are 
                                         * encountered *)
  let allBreaks : brk list ref = ref [] in


     (* Scan the document and initialize the allBreaks data structure. Carry 
      * a stack of aligns and a stack of breakData that are in scope for the 
      * current line. Also carry the current column and a continuation. *)
  let rec scan aligns breaks col cont = function 
      Nil -> cont aligns breaks col
    | Concat(d1,d2) -> 
        scan aligns breaks col (fun a b c -> scan a b c cont d2) d1
    | Text s -> 
        let len = String.length s in
        cont aligns breaks (col + len)
    | CText (d,s) -> 
        scan aligns breaks col 
          (fun a b c -> cont a b (c + String.length s)) d
    | (Break | Line) as doc ->            
        let ar =                        (* Find the matching align. There 
                                         * must always be one because we 
                                         * start with an align in the column 
                                         * 0  *)
          match aligns with
            ar :: _ -> ar
          | _ -> failwith "Bug. Missing align"
        in
                                        (* See if it's time to switch 
                                         * breakAllMode off *)
        let _ = 
          if !breakAllMode && col < width then
            breakAllMode := false else () in
          
                                        (* See if we need to perform some 
                                         * breaks on what we have so far  *)
        let newCol = 
          if !breakAllMode then col else
          performBreaks breaks col  in
                                        (* Maybe we need to switch breakAll 
                                         * mode on *)
        let _ = if newCol > width then breakAllMode := true else () in

                                        (* Create a new breakData *)
        let bid    = incr breakId; !breakId in
        let bdata  = 
          if doc = Break && not !breakAllMode then
            {id=bid;col=newCol + 1;align=ar;status=Open}
          else
            begin
              markNotTaken ar breaks;
              {id=bid;col= !ar      ;align=ar;status=Taken}
            end
        in
                                        (* Memorize it *)
        let _ = allBreaks := bdata :: (!allBreaks) in
        cont aligns (Brk bdata :: breaks) bdata.col

    | Align -> 
        let adata = ref col in
        cont (adata :: aligns) (Algn adata :: breaks) col

    | Unalign -> 
        match aligns with
          _ :: t when t != [] -> cont t breaks col
        | _ -> failwith "Unmatched unalign\n"

  (* scan *)
  in
  let indentLen = ref 0 in
  let doString acc s = 
    let ind = !indentLen in
    let acc' = if ind > 0 then begin
      indentLen := 0; accString acc (String.make ind ' ') end else acc in
    accString acc' s
  in
  let rec layout acc = function
      Nil -> acc
    | Align -> acc
    | Unalign -> acc
    | Text s -> doString acc s
    | Concat (d1, d2) -> layout (layout acc d1) d2
    | CText (d, s) -> doString (layout acc d) s
    | Break | Line -> 
        let b = 
          match !allBreaks with
            b :: t -> allBreaks := t; b
          | [] -> failwith "Pretty: allbreaks list is too short"
        in
        if b.status = Taken then begin
          let _ = if debug then 
            fprintf "taking branch %d to %d\n" b.id (!(b.align)) else () in
          indentLen := !(b.align);
          accString acc "\n"
        end else
          doString acc " "
  in
  function doc -> 
    let scanCont _ _ _ =  
      allBreaks := List.rev (!allBreaks);
      layout start doc
    in
    scan [ref 0] [] 0 scanCont doc



let fprint chn width doc = 
  fit () (fun _ s -> output_string chn s) width doc

let sprint width doc =
  fit "" (fun acc s -> acc ^ s) width doc

let gprint = fit

                                        (* The rest is based on printf.ml *)
external format_int: string -> int -> string = "caml_format_int"
external format_float: string -> float -> string = "caml_format_float"


(* Keep track of nested align in gprintf. Each gprintf format string must 
 * have properly nested align/unalign pairs. When the nesting depth surpasses 
 * !printDepth then we print ... and we skip until the matching unalign *)
let printDepth = ref 100 (* WRW: must see whole proof! *)
let alignDepth = ref 0
    
let gprintf (finish : doc -> doc)  
    (format : ('a, unit, doc) format) : 'a =
  let format = (Obj.magic format : string) in
  
  (* Record the starting align depth *)
  let startAlignDepth = !alignDepth in
  (* Special concatenation functios *)
  let dconcat acc thunk = 
    if !alignDepth > !printDepth then acc else Concat(acc, thunk()) in
  let dctext acc thunk = 
    if !alignDepth > !printDepth then acc else CText(acc, thunk()) in
  (* Special finish function *)
  let dfinish dc = 
    if !alignDepth <> startAlignDepth then
      prerr_string ("Unmatched align/unalign in " ^ format ^ "\n");
    finish dc
  in
  let flen    = String.length format in
                                        (* Reading a format character *)
  let fget    = String.unsafe_get format in
                                        (* Output a literal sequence of 
                                         * characters, starting at i. The 
                                         * character at i does not need to be 
                                         * checked.  *)
  let rec literal acc i = 
    let rec skipChars j = 
      if j >= flen || 
      (match fget j with 
        '%' -> true 
      | '@' -> true 
      | _ -> false) then
        collect (dconcat acc (fun () -> Text (String.sub format i (j-i)))) j
      else
        skipChars (succ j)
    in
    skipChars (succ i)
                                        (* the main collection function *)
  and collect (acc : doc) i = 
    if i >= flen then Obj.magic (dfinish acc) else begin
      let c = fget i in
      if c = '%' then begin
        let j = skip_args (succ i) in
        match fget j with
          '%' -> literal acc j 
        | 's' ->
            Obj.magic(fun s ->
              let thunk () = 
                if j <= i+1 then
                  s
                else
                  let p =
                    try
                      int_of_string (String.sub format (i+1) (j-i-1))
                    with _ ->
                      invalid_arg "fprintf: bad %s format" in
                  if p > 0 && String.length s < p then
                    (String.make (p - String.length s) ' ') ^ s
                  else if p < 0 && String.length s < -p then
                    s ^ (String.make (-p - String.length s) ' ')
                  else
                    s
              in
              collect (dctext acc thunk) (succ j))
        | 'c' ->
            Obj.magic(fun c ->
              collect (dctext acc (fun () -> String.make 1 c)) (succ j))
        | 'd' | 'i' | 'o' | 'x' | 'X' | 'u' ->
            Obj.magic(fun n ->
              collect (dctext acc
                         (fun () -> format_int (String.sub format i 
                                                  (j-i+1)) n))
                (succ j))
        | 'f' | 'e' | 'E' | 'g' | 'G' ->
            Obj.magic(fun f ->
              collect (dctext acc
                         (fun () -> 
                           format_float (String.sub format i (j-i+1)) f))
                (succ j))
        | 'b' ->
            Obj.magic(fun b ->
              collect (dctext acc (fun () -> string_of_bool b)) (succ j))
        | 'a' ->
            Obj.magic(fun pprinter arg ->
              collect (dconcat acc (fun () -> pprinter () arg)) (succ j))
        | 't' ->
            Obj.magic(fun pprinter ->
              collect (dconcat acc pprinter) (succ j))
        | c ->
            invalid_arg ("dprintf: unknown format %s" ^ String.make 1 c)

      end else if c = '@' then begin
        if i + 1 < flen then begin
          match fget (succ i) with

                                        (* Now the special format characters *)
            '[' ->                      (* align *)
              let newacc = 
                if !alignDepth > !printDepth then
                  acc
                else if !alignDepth = !printDepth then
                  CText(acc, "...")
                else
                  Concat(acc, align)
              in
              incr alignDepth;
              collect newacc (i + 2)
                
          | ']' ->                        (* unalign *)
              decr alignDepth;
              let newacc = 
                if !alignDepth >= !printDepth then
                  acc
                else
                  Concat(acc, unalign)
              in
              collect newacc (i + 2)
          | '!' ->                        (* hard-line break *)
              collect (dconcat acc (fun () -> line)) (i + 2)
          | '?' ->                        (* soft line break *)
              collect (dconcat acc (fun () -> break)) (i + 2)
          | '@' -> 
              collect (dctext acc (fun () -> "@")) (i + 2)
          | c ->
              invalid_arg ("dprintf: unknown format @" ^ String.make 1 c)
        end else
          invalid_arg "dprintf: incomplete format @"
      end else if c = '\\' then begin
        if i + 1 < flen then begin
          match fget (succ i) with
            'n' ->                      
              collect (dconcat acc (fun () -> line)) (i + 2)
          | _ -> literal acc i
        end else
          invalid_arg "dprintf: incomplete escape \\"
      end else
        literal acc i
    end

  and skip_args j =
    match String.unsafe_get format j with
      '0' .. '9' | ' ' | '.' | '-' -> skip_args (succ j)
    | c -> j

  in
  collect Nil 0

let dprintf format     = gprintf (fun x -> x) format
let fprintf chn format = 
  let f d = fprint chn 80 d; d in
	(* weimeric hack begins -- flush output to streams *)
	let res = gprintf f format in
	(* save the value we would have returned, flush the channel and then
	 * return it -- this allows us to see debug input near infinite loops *)
	ignore (flush chn) ; res
	(* weimeric hack ends *)
	

let printf format = fprintf stdout format
let eprintf format = fprintf stderr format

(*
let sprintf format = gprintf (fun x -> sprint 80 x) format
let printf  format = gprintf (fun x -> fprint stdout 80 x) format
*)

